#!@@PERL@@
#
# This program is used to display information collected by the spong server.
# This is a CGI program and expects to be run as such.  You can either place
# this in your web server's cgi-bin directory, or set up a link.  Note that
# your status, history, and system host information will need to be readable
# by the same user that your web server runs as (typically nobody).
#
# History:
# (1) Ported mkbb.* scripts to perl. (Ed Hill Mar 1, 1997)

use Sys::Hostname;
use POSIX;

# Load our configuration variables, including anything specific to the host
# that we are running on.

$|++;
$conf_file  = "@@ETCDIR@@/spong.conf";
$hosts_file = "@@ETCDIR@@/spong.hosts";
$screens    = "@@WWWDIR@@/screens";
$cmd        = $ENV{'PATH_INFO'};
($HOST)     = gethostbyname(&Sys::Hostname::hostname());
$HOST       =~ tr/A-Z/a-z/;

# Figure out who I am and how I am being called.

if( $ENV{'HTTPS'} eq "on" ) {$script = "https://";} else {$script = "http://";}
$script .= $ENV{'SERVER_NAME'};
if( $ENV{'SERVER_PORT'} ne "80" && $ENV{'SERVER_PORT'} ne "443" ) {
   $script .= ":" . $ENV{'SERVER_PORT'};
}
$script .= $ENV{'SCRIPT_NAME'};
($webdir) = ($script =~ /^(.*)\/spong-display.cgi.*$/);
$imagedir = "$webdir/gifs";

require $conf_file || die "Can't load $conf_file: $!";
if( -f "$conf_file.$HOST" ) {
   require "$conf_file.$HOST" || die "Can't load $conf_file.$HOST: $!";
} else {
   my $tmp = (split( /\./, $HOST ))[0];
   if( -f "$conf_file.$tmp" ) { # for lazy typist
      require "$conf_file.$tmp" || die "Can't load $conf_file.$tmp: $!";
   }
}

# Read in the spong.hosts file.  We are a little nasty here in that we do some
# junk to scan through the file so that we can maintain the order of the hosts
# as they appear in the file.

open( HOSTS, $hosts_file ) || die "Can't load $hosts_file: $!";
while( <HOSTS> ) {
   $evalme .= $_;
   if( /^\s*%HOSTS\s*=\s*\(/ ) { $inhosts = 1; }
   if( $inhosts && /^\s*[\'\"]?([^\s\'\"]+)[\'\"]?\s*\=\>\s*\{/ ) {
      push( @HOSTS_LIST, $1 );
   }
}
close( HOSTS );
eval $evalme || die "Invalid spong.hosts file: $@";

# Fallback, if we didn't read things correctly...
if( sort @HOSTS_LIST != sort keys %HOSTS ) { @HOSTS_LIST = sort keys %HOSTS; }


# Process any options that we are given..

if( $cmd eq "" || $cmd eq "/" ) { &show( &template( "home" ) );    exit(0); }

if( $cmd =~ /^\/([^\/]+)$/ )    { &show( &template( $1 ) );        exit(0); }
if( $cmd =~ /^\/help\/(.*)$/ )  { &show( &template( "help-$1" ) ); exit(0); }
if( $cmd =~ /^\/host\/(.*)$/ )  { &host( $1 );                     exit(0); }

if( $cmd =~ /^\/service\/(.*?)\/(.*)$/ ) { &service( $1, $2 );     exit(0); }

exit(0);


# This takes a web page, replaces all the templates, and sends it to stdout...

sub show {  
   my( $template ) = @_;
   my( $key, $val, %replace );
   my $expire_date = localtime();
   my $update      = POSIX::strftime( "%H:%M, on %D", localtime() );
   my $errors;

   # First replace the "simple" stuff...
   
   $template =~ s/__UPDATE_TIME__/$update/g;
   $template =~ s/__RELOAD__/$SPONGSLEEP/;
   $template =~ s/__EXPIRE_DATE__/$expire_date/g;
   $template =~ s/__SCRIPT__/$script/g;
   $template =~ s/__IMAGES__/$imagedir/g;
   $template =~ s/__DIR__/$webdir/g;

   if( $template =~ m/__HISTORY_ALL__/g ) {
      my $history = History->new()->display( 7 );
      $template =~ s/__HISTORY_ALL__/$history/isg;
   }

   # First pass, go through and build a replacement hash.  For each of our
   # more "complicated" templates, figure out what the template wants and
   # try to generate some data for it...

   while( $template =~ m/__GROUP_(.*?)_(.*?)__/isg ) {
      my( $groupstr, $format ) = ( $1, $2 );
      
      next if $replace{"GROUP_${groupstr}_$format"};

      my $group = Group->new( $groupstr );
      $replace{"GROUP_${groupstr}_$format"} = $group->display( $format );
   }

   # Now go through the hash, and actually do the replacements.

   while( ($key, $val) = each( %replace ) ) {
       $template =~ s/(__\Q$key\E__)/$val/gs; }

   # This is kind of (ok it is really) a hack...  Perhaps some day I'll clean
   # this up...

   my $message;
   if( $template =~ /__STALE_DATA__/ ) {
      $message = "<font color=green><b>No known problems (*)</b></font>";
      $message .= "<p><font size=-1 color=purple>Some information is out of ";
      $message .= "date which might indicate a problem.</font>";
   } else {
      $message = "<font color=green><b>No current problems</b></font>";
   }

   $template =~ s/(__STALE_DATA__)//isg;
   $template =~ s/(__HACK_START__\s*__HACK_END__)/$message/isg;
   $template =~ s/(__HACK_START__)//isg;
   $template =~ s/(__HACK_END__)//isg;

   print "Content-type: text/html\n\n", $template;
}

# Read a file name in the screens directory, and build a string that will get
# passed to &show() so that it's template tag can be translated to the real
# stuff...

sub template {
   my( $file ) = @_;
   my $template;

   if( -f "$main::screens/$file.html" ) {
      open( TEMPLATE, "$main::screens/$file.html" ) || die "Ack... $!";
      while( <TEMPLATE> ) { $template .= $_; };
      close( TEMPLATE );
   } else {
      print "Content-type:text/html\n\n";
      print "<h1>Error - Request not understood!</h1>\n";
      exit(0);
   }

   return $template;
}

# This shows the current information about a specific host.  This would 
# provide more detail about the problem, as well as provide history, and 
# anything else we would want to include.

sub host { &show( Host->new( $_[0] )->display() ); }

# This shows complete information about a specific message.  This would 
# provide all detail available about the problem.  It would also provide
# a history of problems with a specific service.

sub service { &show( Service->new( @_ )->display() ); }


# ===========================================================================
# That's it of the CGI part of this file.  Below here are 3 class definitions
# and the HTML templates used to generate the UI elements.  The classes are:
#
#   * Group     Knows what hosts are in a group, loads them, displays summary
#   * Host      Knows what services run on a given host, displays host info
#   * Service   Knows status of a specific service on a host, displays detail
#   * History   Knows about problems that have happened over time
# ===========================================================================

package Service;

%Service::names = {};

sub names {
   my( $class ) = @_;
   my( $service, @services );

   # Always return ping first.  This probably isn't the place to build this
   # smarts, but I'm going to for now...

   if( grep( /^ping$/, keys %Service::names ) ) { push( @services, "ping" ); }
   foreach $service ( sort keys %Service::names ) {
      push( @services, $service ) unless $service eq "ping";
   }
   
   return @services;
}

sub new {
   my( $class, $host, $service, $color ) = @_;
   my( $self, $mtime, $htime ) = ( {}, "", "" );

   $Service::names{$service} = 1;

   # At this level there are only three possible colors (red, yellow, green)
   # so only check for those stats, we compute other colors (purple) based
   # on other information

   if( -d $main::SPONGSTATUS ) {
      if( $color ne "" ) {
	 if( -f "$main::SPONGSTATUS/$host-$service-$color" ) { 
	    $mtime = (stat(_))[9]; }
      } else {
	 if( -f "$main::SPONGSTATUS/$host-$service-green" ) { 
	    $mtime = (stat(_))[9]; $color = "green"; }
	 if( -f "$main::SPONGSTATUS/$host-$service-yellow" ) { 
	    $mtime = (stat(_))[9]; $color = "yellow"; }
	 if( -f "$main::SPONGSTATUS/$host-$service-red" ) { 
	    $mtime = (stat(_))[9]; $color = "red"; }
      }
   }

   # Check to see when we were last updated, if we have not been updated
   # in a while, change our color to purple.

   $self->{'lcolor'} = $color;
   if( $mtime < ( time() - (3 * $main::SPONGSLEEP) ) ) { $color = "purple"; }
      
   # Check to see when the history for this service was last updated.

   if( -d $main::SPONGHISTORY ) {
      if( -f "$main::SPONGHISTORY/$host-$service" ) { 
	 $htime = (stat(_))[9]; 
      }
   }

   $self->{'name'}  = $service;
   $self->{'host'}  = $host;
   $self->{'color'} = $color;
   $self->{'mtime'} = $mtime;
   $self->{'htime'} = $htime;

   bless $self;
   return $self;
}

sub name   { return $_[0]->{'name'}; }
sub host   { return $_[0]->{'host'}; }
sub color  { return $_[0]->{'color'}; }
sub lcolor { return $_[0]->{'lcolor'} }
sub mtime  { return $_[0]->{'mtime'}; }
sub htime  { return $_[0]->{'htime'}; }

sub summary {
   my( $self ) = @_;
   my $file    = $self->host() . "-" . $self->name() . "-" . $self->lcolor();
   my $str     = "";

   open( FILE, "$main::SPONGSTATUS/$file" );
   my $summary = <FILE>; chomp $summary;
   close( FILE );
   my( $time, $rest ) = ( $summary =~ /^(\d+) (.*)$/ );

   return substr( $rest, 0, 50 );
}

sub message {
   my( $self ) = @_;
   my $file    = $self->host() . "-" . $self->name() . "-" . $self->lcolor();
   my $str     = "";

   open( FILE, "$main::SPONGSTATUS/$file" );
   my $summary = <FILE>; chomp $summary;
   my( $time, $rest ) = ( $summary =~ /^(\d+) (.*)$/ );

   $str = POSIX::strftime( "%H:%M, %D", localtime($time) );
   $str .= "  $rest\n\n";
   
   while( <FILE> ) { $str .= $_; }
   close( FILE );

   return $str;
}

   
sub display { 
   my( $self )  = @_;
   my $template = "";
   my $color    = $self->color();
   my $col    = "<img src=\"$main::imagedir/$color.gif\" alt=$color border=0>";
   my $host     = $self->host();
   my $service  = $self->name();
   my $history  = History->new( $host, $service )->display( 7 );
   my $message  = $self->message();
   my $expire   = localtime();

   open( TEMPLATE, "$main::screens/service.html" ) || die "Ack... $!";
   while( <TEMPLATE> ) { $template .= $_; };
   close( TEMPLATE );

   $template =~ s/__RELOAD__/$main::SPONGSLEEP/;
   $template =~ s/__EXPIRE_DATE__/$expire/g;

   $template =~ s/__HOST__/$host/isg;
   $template =~ s/__IMAGES__/$main::imagedir/isg;
   $template =~ s/__COLOR__/$col/isg;
   $template =~ s/__SERVICE__/$service/isg;
   $template =~ s/__HISTORY__/$history/isg;
   $template =~ s/__MESSAGE__/$message/isg;

   return $template;
}


# ===========================================================================

package Host;

%Host::names = {};

sub names {
   my( $class ) = @_;
   my( $host, @hosts );

   foreach $host ( sort keys %Host::names ) { push( @hosts, $host ); }
   
   return @hosts;
}

sub new {
   my( $class, $host ) = @_;
   my $self = {};
   my( %servicehash );

   # Go through the status directory and for each service we find there
   # associated with this host, add it to our service list.

   if( -d $main::SPONGSTATUS ) {
      opendir( DIR, $main::SPONGSTATUS ) || 
	 die "Can't read $main::SPONGSTATUS: $!";
      while( $_ = readdir( DIR ) ) {
	 if( /^$host\-([^\-]+)\-([^\-]+)$/ ) {
	    my( $service, $color ) = ($1, $2);
	    my $object = Service->new( $host, $service, $color );

	    $servicehash{$service} = $object;
	 }
      }
      closedir( DIR );
   }

   $self->{'name'}          = $host;
   $self->{'servicehash'}   = \%servicehash;
   $self->{'services'}      = [ values %servicehash ];
   $self->{'service_names'} = [ sort keys %servicehash ];

   bless $self;
   return $self;
}

sub name          { return $_[0]->{'name'}; };
sub service       { return $_[0]->{'servicehash'}->{$_[1]}; }
sub services      { return @{$_[0]->{'services'}}; };
sub service_names { return @{$_[0]->{'service_names'}}; };

sub problem {
   my( $self ) = @_;
   my $name = $self->name();
   my $problem;

   foreach( $self->services() ) { 
      if( $_ && ($_->color() eq "red") ) { $problem = "red"; }
      if( $_ && ($_->color() eq "purple") ) {
	 $problem = "purple" unless $problem eq "red"; }
   }

   # If there is a problem reported, and we are in this machine's down time
   # then don't report it.

   foreach( @{$main::HOSTS{$name}->{'down'}} ) {
      my( $day, $shour, $smin, $ehour, $emin ) = 
	 ( /^(.):(\d+):(\d+)-(\d+):(\d+)$/ );
      my( $nday, $nhour, $nmin ) = (localtime())[6,2,1];

      if( $day eq "*" || $nday eq $day ) {
	 my $ntotal = $nhour * 60 + $nmin;
	 my $stotal = $shour * 60 + $smin;
	 my $etotal = $ehour * 60 + $emin;

	 if( $ntotal >= $stotal && $ntotal <= $etotal ) {
	    $problem = "";
	 }
      }
   }
   
   return $problem;
}

sub display { 
   my( $self )     = @_;
   my $host        = $self->name();
   my $expire_date = localtime();
   my $update      = POSIX::strftime( "%H:%M on %D", localtime() );
   my $history     = History->new( $host, "" )->display( 7 );
   my $expire      = localtime();
   my $template;
   my( $table, $service );

   open( TEMPLATE, "$main::screens/host.html" ) || die "Ack... $!";
   while( <TEMPLATE> ) { $template .= $_; };
   close( TEMPLATE );

   $template =~ s/__RELOAD__/$main::SPONGSLEEP/;
   $template =~ s/__EXPIRE_DATE__/$expire/g;
   $template =~ s/__HOST__/$host/isg;
   $template =~ s/__UPDATE_TIME__/$update/isg;
   $template =~ s/__HISTORY__/$history/isg;
   $template =~ s/__IMAGES__/$main::imagedir/isg;

   if( -f "$main::SPONGNOTES/$host" ) {
      open( FILE, "$main::SPONGNOTES/$host" );
      while( <FILE> ) { $information .= $_; }
      close( FILE );
   }
	 
   $table = "<table width=100% border=1 cellspacing=2 cellpadding=2><tr>\n";
   $table .= "<td width=10% align=center><b>Service</b></td>";
   $table .= "<td width=1%>&nbsp</td>";
   $table .= "<td width=25% align=center><b>Updated</b></td>";
   $table .= "<td width=65% align=center><b>Summary</b></td>";
   $table .= "</tr>\n";
   
   foreach $service ( $self->services() ) {
      $color = $service->color();
      $table .= "<tr><td align=left valign=top>";
      $table .= "<a href=\"$main::script/service/$host/";
      $table .= $service->name() . "\">" . $service->name() . "</a></td>";
      $table .= "<td align=center valign=top>";
      $table .= "<img src=\"$main::imagedir/$color.gif\" alt=$color border=0>";
      $table .= "</td><td align=center valign=top>";
      $table .= POSIX::strftime( "%H:%M, %D", localtime($service->mtime()) );
      $table .= "</td><td align=left valign=top>" . $service->summary();
      $table .= "</td></tr>\n";
   }
   $table .= "</table>\n";

   $template =~ s/__SERVICE_TABLE__/$table/isg;
   $template =~ s/__INFORMATION__/$information/isg;
   
   return $template;
}

sub display_problem {
   my( $self ) = @_;
   my( $service, @problist, $probtime, $list );
   my $spacer = "&nbsp &nbsp &nbsp ";
   my $host = $self->name();

   foreach $service ( $self->services() ) {
      if( $service && $service->color() eq "red" ) {
	 push( @problist, $service->name() );
	 if( $probtime == 0 || $service->htime() < $probtime ) {
	    $probtime = $service->htime();
	 }
      }
   }

   $list .= "<image src=$main::imagedir/red.gif alt=red border=0> ";
   $list .= "<b><a href=\"$main::script/host/$host\">$host</a></b><br> ";

   if( $#problist > 0 ) {
      $list .= "$spacer problem: multiple prob.<br>";
   } else {
      $list .= "$spacer problem: <a href=\"$main::script/service/$host/";
      $list .= $problist[0] . "\">" . $problist[0] . "</a><br>";
   }

   if( $probtime > 0 ) {
      $time = POSIX::strftime( "%H:%M, %D", localtime($probtime) );
      $list .= "$spacer time: $time<br>";
   }

   my $contact = $main::HOSTS{$host}->{'contact'};
   my $email   = $main::HUMANS{$contact}->{'email'};

   $list .= "$spacer contact: ";
   if( $email ) { $list .= "<a href=\"mailto:$email\">"; }
   $list .= $main::HUMANS{$contact}->{'name'};
   if( $email ) { $list .= "</a>"; }
   $list .= "<p>\n";

   return $list;
}


# ===========================================================================

package Group;

sub new {
   my( $class, $group ) = @_;
   my $self = {};
   my( %hosthash, @names, $s, %slist, @hostnames );

   # First load all the hosts for our group.

   foreach $host ( @main::HOSTS_LIST ) {
      my $groupname = $main::HOSTS{$host}->{'group'};

      if( $groupname eq $group || $group eq "ALL" ) {
	 my $object;
	 if( $main::hostcache{$host} ne "" ) {
	    $object = $main::hostcache{$host};
	 } else {
	    $object = Host->new( $host );
	    $main::hostcache{$host} = $object;
	 }

	 $hosthash{$host} = $object;
	 push( @hostnames, $host );

	 foreach $s ( $object->service_names() ) { $slist{$s} = 1; } 
      }
   }

   $self->{'name'}          = $group;
   $self->{'hosthash'}      = \%hosthash;
   $self->{'service_names'} = [ sort keys %slist ];
   $self->{'host_names'}    = [ @hostnames ];
   $self->{'hosts'}         = [];

   # Keep the hosts in the same order they were read in as.

   foreach( @hostnames ) {
      push( @{$self->{'hosts'}}, $hosthash{$_} ); }

   bless $self;
   return $self;
}

sub name          { return $_[0]->{'name'}; }
sub host          { return $_[0]->{'hosthash'}->{$_[1]}; }
sub hosts         { return @{$_[0]->{'hosts'}}; }
sub host_names    { return @{$_[0]->{'host_names'}}; }
sub service_names { return @{$_[0]->{'service_names'}}; }

sub service {
   my( $self, $hostname, $servicename ) = @_;
   my( $hostobj ) = $self->host( $hostname );
   my( $servobj ) = $hostobj->service( $servicename );

   return $servobj;
}
   
   
   
# This generates the master table for a series of hosts.  This is a 2 
# dimensional table with hosts down the vertical axis, and services across
# the top axis, the contents being the current state of that service on that
# machine.  You can click on the hosts (or the gif indicators inside the 
# table) to show more information about that host.  You can also click on 
# the service names to show just a small text explanation of that service.

sub display { 
   my( $self, $type ) = @_;

   if( $type eq "ALL" ) {
      my( $service, $table, $host );

      $table = "<table border=1><tr><td align=center width=80>";
      $table .= "<b>Host</b></td>\n";

      foreach( Service->names() ) { 
	 $table .= "<td align=center valign=bottom width=20>\n";
	 $table .= "<font size=-1><a href=\"$main::script/help/$_\">";
	 $table .= "$_</a></font></td>\n"; 
      }
      $table .= "</tr>\n\n";

      foreach $host ( $self->host_names() ) { 
	 my $short = (split( /\./, $host ))[0];
	 $table .= "<tr><td align=left>\n";
	 $table .= "<a href=\"$main::script/host/$host\">$short</a></td>\n";

	 foreach $servicename ( Service->names() ) { 
	    my $service = $self->service( $host, $servicename );

	    if( $service ne "" ) {
	       my $col = $service->color();
	       $table .= "<td align=center><a href=\"$main::script/service/";
	       $table .= "$host/$servicename\">";
	       $table .= "<image src=$main::imagedir/$col.gif alt=$col ";
	       $table .= "border=0></a></td>\n"; 
	    } else {
	       $table .= "<td align=center> - </td>\n"; 
	    }
	 }
	 $table .= "</tr>\n\n";
      }
      $table .= "</table>\n";
      return $table;
   } elsif( $type eq "ERRORS" ) {
      my( $list, $host, $somestale );

      foreach $host ( $self->hosts() ) { 
	 my $problem = $host->problem();
	 if( $problem eq "red" ) { $list .= $host->display_problem(); }
	 if( $problem eq "purple" ) { $somestale = 1; }
      }

      if( $list eq "" && $somestale ) { $list .= "__STALE_DATA__"; };

      return $list;
   }

}

# ===========================================================================

package History;

sub new {
   my( $class, $host, $service ) = @_;
   my $self = {};
   my( @history );

   # Scan through the history directory, and read in all the files that are
   # applicable.

   if( -d $main::SPONGHISTORY ) {
      opendir( DIR, $main::SPONGHISTORY ) || 
	 die "Can't read $main::SPONGHISTORY: $!";
      while( $_ = readdir( DIR ) ) {
	 if( /^(.+)\-([^\-]+)$/ ) {
	    my( $hhost, $hservice ) = ($1, $2);

	    if( $host ne "" && $hhost ne $host ) { next; }
	    if( $service ne "" && $hservice ne $service ) { next; }

	    open( FILE, "$main::SPONGHISTORY/$hhost-$hservice" ) ||
	       die "Can't open $main::SPONGHISTORY/$hhost-$hservice: $!";
	    while( <FILE> ) {
	       chomp;
	       my( $time, $color, $summary) = ( /^(\d+)\s+(\S+)\s+(.*)$/ );
	       push( @history, "$time $hhost $hservice $color $summary" );
	    }
	    close( FILE );
	 }
      }
   }

   $self->{'history'} = [ reverse sort @history ];

   # Keep the hosts in the same order they were read in as.

   bless $self;
   return $self;
}

sub history { return @{$_[0]->{'history'}}; }
   
sub display {
   my( $self, $ndays ) = @_;
   my( $str, $date, $intable );

   foreach( $self->history() ) {
      my( $time, $host, $service, $color, $summary ) = 
	 ( /^(\d+)\s+(\S+)\s+(\w+)\s+(\w+)\s+(.*)$/ );
      next if $color eq "";
      next if( $time < (time() - ($ndays * 24*60*60)) );

      my $shost = (split( /\./, $host ))[0];
      $dstr = POSIX::strftime( "%A, %D", localtime($time) );

      if( $dstr ne $date ) {
	 $date = $dstr;
	 if( $intable ) { $str .= "</table><p>\n"; }

	 $str .= "$dstr<hr border=1 noshade>\n";
	 $str .= "<table width=100% border=0 cellspacing=0 cellpadding=0>\n";
      }

      $intable = 1;
      $str .= "<tr><td width=2% align=center valign=top>";
      $str .= "<img src=\"$main::imagedir/$color.gif\" alt=$color border=0>";
      $str .= "</td>\n";
      $str .= "<td width=10% align=center valign=top>";
      $str .= sprintf( "%2.2d:%2.2d", (localtime($time))[2,1] );
      $str .= "</td>\n";
      $str .= "<td width=18% align=left valign=top>$shost </td>\n";
      $str .= "<td width=10% align=left valign=top>$service</td>\n";
      $str .= "<td align=left valign=top>$summary</td>\n</tr>\n";
   }

   $str .= "</table>\n";

   return $str;
}
